
#include <libc.h>
#include <i86.h>
#include <sys/inline.h>
#include <sys/seginfo.h>
#include <sys/name.h>
#include "nmi.h"


typedef struct i86_gated i86_gated;
typedef struct i86_segd i86_segd;
#pragma pack(1)
struct i86_gated {
	ushort_t        offs_low;
	ushort_t        selector;
	uchar_t         argcount;
	uchar_t         type;
	ushort_t        offs_high;
};
struct i86_segd {
	ushort_t        limlo;
	ushort_t        baselo;
	uchar_t         basehi;
	uchar_t         stype;
	uchar_t         limflags;
	uchar_t         basexhi;
};
#pragma pack();
#define isgdt(x)      (((unsigned)(x)&4)==0)
#define seltoi(x)     ((unsigned)(x)>>3)

#define IDT ((i86_gated far*)MK_FP(0x10, 0))
#define GDT ((i86_segd  far*)MK_FP(0x8, 0))

#define NMI_VEC 2

static
mk_igate(i86_gated * g, ulong_t f, unsigned s)
{
	g->offs_low = (ushort_t) f;
	g->offs_high = (ulong_t) f >> 16;
	g->selector = s;
	g->argcount = 0;
	g->type = 0xee;
}



static void
setvec(int vec, i86_gated * ngate, i86_gated * ogate)
{
	int             ps;
	ps = disable();
	if (ogate)
		*ogate = IDT[vec];
	IDT[vec] = *ngate;
	restore(ps);
}

static int
setpl(unsigned segno, unsigned npl)
{
	uchar_t         otype;

	if (npl > 3 || !isgdt(segno))
		return -1;
	otype = GDT[seltoi(segno)].stype;
	GDT[seltoi(segno)].stype = otype & ~(3 << 5) | npl << 5;
	return otype >> 5 & 3;
}




static unsigned gdt_ds, gdt_cs;
static          ogate_set;
static i86_gated oldgate;
static unsigned name_id;
extern unsigned short nmi_dseg;

void
uninstall_nmi(int signo)
{
	if (ogate_set) {
		setvec(NMI_VEC, &oldgate, 0);
		ogate_set = 0;
	}
	if (gdt_cs) {
		qnx_segment_free(gdt_cs);
		gdt_cs = 0;
	}
	if (gdt_ds) {
		qnx_segment_free(gdt_ds);
		gdt_ds = 0;
	}
	if (name_id) {
		qnx_name_detach(0, name_id);
		name_id = 0;
	}
	_exit(1);
}


int
install_nmi(char *name, void hdlr(void))
{
	i86_gated       newgate;
	int             i;
	volatile ushort_t *patch = (void *)__nmi_dseg_patch;
#ifdef ISFLAT
#define FLAT 1
#else
#define FLAT 0
#endif

#ifndef _NSIG
#define _NSIG _SIGMAX
#endif


	for (i = 1; i < _NSIG; i++)
		signal(i, uninstall_nmi);

	if ((name_id = qnx_name_attach(0, name)) == -1) {
		return -1;
	}
	_nmi_funcp = (ulong_t) hdlr;

	if ((gdt_ds = qnx_segment_put(0, _ds(), _PMF_MODIFY | _PMF_GLOBAL)) == -1) {
		uninstall_nmi(0);
		return -1;
	}
	gdt_ds &= ~3;

	if (!FLAT) {
		struct _seginfo buf;
		unsigned        temp;
		if (qnx_segment_info(0, 0, _cs(), &buf) == -1) {
			uninstall_nmi(0);
			return -1;
		}
		if ((temp = qnx_segment_overlay_flags(buf.addr, buf.nbytes,
				     _PMF_DATA_RW | _PMF_VOVERLAY)) == -1) {
			uninstall_nmi(0);
			return -1;
		}
		*(ushort_t far *) MK_FP(temp, (ulong_t)patch) = gdt_ds;
		qnx_segment_free(temp);
	}  else {
		*patch = gdt_ds;
	}

	if ((gdt_cs = qnx_segment_put(0, _cs(), _PMF_INUSE | _PMF_CODE_RX | _PMF_GLOBAL | _PMF_MODIFY)) == -1) {
		uninstall_nmi(0);
		return -1;
	}
	gdt_cs &= ~3;
	setpl(gdt_ds, 0);
	setpl(gdt_cs, 0);

	mk_igate(&newgate, (ulong_t) __nmi_handler, gdt_cs);
	printf("gate: 0x%x:0x%x%04.4x\n",
	       newgate.selector, newgate.offs_high, newgate.offs_low);
	setvec(NMI_VEC, &newgate, &oldgate);
	ogate_set = 1;
	return 0;
}
